import argparse
import math
import os
import random
import sys
from boonamber import AmberClient, AmberCloudError


def SignalGenerator(wave='sine', samplerate=1.0, amp=1.0, freq=0.01, phase=0.0, noise=0.0):
    """Producer which generates noisy data from a given waveform.

        Args:
            wave: waveform to use (one of 'sine', 'square', 'triangle')
            samplerate: signal sampling rate
            amp: wave amplitude
            freq: frequency in Hz
            phase: phase offset
            noise: sigma of added Gaussian noise
        
        Produces:
            sequential scalar values from the specified signal
    """

    try:
        sine = lambda x: math.sin(x)
        square = lambda x: math.copysign(1, math.sin(x))
        triangle = lambda x: (2 / math.pi) * math.asin(math.sin(x))
        func = {'sine': sine,
                'square': square,
                'triangle': triangle}[wave]
    except KeyError:
        raise ValueError("unsupported wave: '{}'".format(wave))

    t = 0.0
    while True:
        yield amp * func(2 * math.pi * freq * t + phase) + random.gauss(0.0, noise)
        t += 1 / samplerate


def main():
    signal = SignalGenerator(wave=args.waveform,
                             samplerate=args.sample_rate,
                             amp=args.amp,
                             freq=args.freq,
                             phase=args.phase,
                             noise=args.noise)

    amber = AmberClient(license_file="test.Amber.license")
    amber.password = os.environ['AMBER_TEST_PASSWORD']

    sensor_id = '846f218ab552fa82'
    if args.reset:
        try:
            amber.configure_sensor(sensor_id, feature_count=1, streaming_window_size=25)
        except AmberCloudError as e:
            print(e)
            sys.exit(1)

    try:
        config = amber.get_config(sensor_id)
    except AmberCloudError as e:
        print(e)
        sys.exit(1)
    print("using sensor: {}".format(sensor_id))
    print("configuration: {}".format(config))
    print()

    batch = []
    for sample in signal:
        if random.random() < args.anomaly_prob:
            max_skips = max(1, int((1 / args.freq) / args.sample_rate))
            num_skips = random.randint(1, max_skips)
            for _ in range(num_skips):
                next(signal)

        batch.append(sample)
        if len(batch) < args.batch_size:
            continue

        try:
            results = amber.stream_sensor(sensor_id, batch)
        except AmberCloudError as e:
            print(e)
            sys.exit(1)

        batch = []

        if results['state'] in ["Buffering", "Autotuning"]:
            state_pretty = '{} ({}%)'.format(results['state'], results['progress'])
        else:
            state_pretty = results['state']
        si_pretty = ' '.join("{:4.3f}".format(val) for val in results['SI'])
        print("{}: {} inferences, {} clusters, SI: [{}]".format(
            state_pretty, results['totalInferences'], results['clusterCount'], si_pretty))


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--sample-rate', nargs='?', default=0.1, type=float,
                        help="interval between data samples (sec)")
    parser.add_argument('--batch-size', nargs='?', default=10, type=int,
                        help="data samples per stream request")
    parser.add_argument('--waveform', nargs='?', default='sine', type=str,
                        help="synthetic data waveform (one of 'sine', 'square', 'triangle')")
    parser.add_argument('--amp', nargs='?', default=1.0, type=float,
                        help="synthetic data amplitude")
    parser.add_argument('--freq', nargs='?', default=0.33, type=float,
                        help="synthetic data frequency (Hz)")
    parser.add_argument('--phase', nargs='?', default=0.0, type=float,
                        help="synthetic data phase offset (rad)")
    parser.add_argument('--noise', nargs='?', default=0.1, type=float,
                        help="sigma of added Gaussian noise ")
    parser.add_argument('--anomaly-prob', nargs='?', default=1 / 1200, type=float,
                        help="baseline probability of anomaly occuring at any given sample")
    parser.add_argument('--reset', action='store_true',
                        help="if set, reset sensor at start of script")
    args = parser.parse_args()
    main()



import itertools
import numpy as np
import json
import os
import requests
import time
import requests
from collections.abc import Iterable
from numbers import Number, Integral
from urllib3.exceptions import InsecureRequestWarning


############################
# Boon Amber Python SDK v1 #
############################


class AmberUserError(Exception):
    """Raised to indicate an error in SDK usage"""

    def __init__(self, message):
        self.message = message


class AmberCloudError(Exception):
    """Raised upon any non-200 response from the Amber cloud"""

    def __init__(self, code, message):
        self.code = code
        self.message = message
        super().__init__("{}: {}".format(code, message))


class AmberClient:

    def __init__(self, license_id='default', license_file="~/.Amber.license", verify=True, cert=None, timeout=300):
        """Main client which interfaces with the Amber cloud. Amber account
        credentials are discovered within a .Amber.license file located in the
        home directory, or optionally overridden using environment variables.

        Args:
            license_id (str): license identifier label found within .Amber.license file
            license_file (str): path to .Amber.license file
            verify:  Boolean, controls whether we verify the server’s TLS certificate
            cert (bool): if String, path to ssl client cert file (.pem). If Tuple, (‘cert’, ‘key’) pair.

        Environment:

            `AMBER_LICENSE_FILE`: sets license_file path

            `AMBER_LICENSE_ID`: sets license_id

            `AMBER_USERNAME`: overrides the username as found in .Amber.license file

            `AMBER_PASSWORD`: overrides the password as found in .Amber.license file

            `AMBER_SERVER`: overrides the server as found in .Amber.license file

            `AMBER_OAUTH_SERVER`: overrides the oauth server as found in .Amber.license file

            `AMBER_SSL_CERT`: path to ssl client cert file (.pem)

            `AMBER_SSL_VERIFY`: Either a boolean, in which case it controls whether we verify the server’s TLS certificate, or a string, in which case it must be a path to a CA bundle to use

        Raises:
            AmberUserError: if error supplying authentication credentials
        """

        self.token = None
        self.reauth_time = 0
        self.user_agent = 'Boon Logic / amber-python-sdk / requests'
        self.timeout = timeout

        # first load from license file, override from environment if specified
        self.license_file = license_file
        self.license_file = os.environ.get('AMBER_LICENSE_FILE', self.license_file)

        # determine which license_id to use, override from environment if specified
        self.license_id = license_id
        self.license_id = os.environ.get('AMBER_LICENSE_ID', self.license_id)

        # create license profile
        if self.license_file is not None:
            license_path = os.path.expanduser(self.license_file)
            if not os.path.exists(license_path):
                self.license_profile = json.loads('{"username": "", "password": "", "server": "", "oauth-server": ""}')
            else:
                try:
                    with open(license_path, 'r') as f:
                        file_data = json.load(f)
                except json.JSONDecodeError as e:
                    raise AmberUserError(
                        "JSON formatting error in license file: {}, line: {}, col: {}".format(e.msg, e.lineno, e.colno))
                try:
                    self.license_profile = file_data[self.license_id]
                except KeyError:
                    raise AmberUserError("license_id \"{}\" not found in license file".format(self.license_id))
        else:
            # no license file found, create a stub profile to be filled in from environment
            self.license_profile = json.loads('{"username": "", "password": "", "server": "", "oauth-server": ""}')

        # override from environment if specified
        try:
            self.license_profile['username'] = os.environ.get('AMBER_USERNAME', self.license_profile['username'])
            self.license_profile['password'] = os.environ.get('AMBER_PASSWORD', self.license_profile['password'])
            self.license_profile['server'] = os.environ.get('AMBER_SERVER', self.license_profile['server'])
            if 'oauth-server' not in self.license_profile:
                self.license_profile['oauth-server'] = self.license_profile['server']
            self.license_profile['oauth-server'] = os.environ.get('AMBER_OAUTH_SERVER', self.license_profile['oauth-server'])
            self.license_profile['cert'] = os.environ.get('AMBER_SSL_CERT', cert)
            verify_str = os.environ.get('AMBER_SSL_VERIFY', "true").lower()
            self.license_profile['verify'] = True  # Default
            if not verify or verify_str == "false":
                self.license_profile['verify'] = False
        except KeyError as e:
            raise AmberUserError("missing field")

        if self.license_profile['verify'] is False:
            requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)

        # verify required profile elements have been created
        if self.license_profile['username'] == "":
            raise AmberUserError("username \"{}\" not specified".format(self.license_profile['username']))
        if self.license_profile['password'] == "":
            raise AmberUserError("password \"{}\" not specified".format(self.license_profile['password']))
        if self.license_profile['server'] == "":
            raise AmberUserError("server \"{}\" not specified".format(self.license_profile['server']))

    def _authenticate(self):
        """Authenticate client for the next hour using the credentials given at
        initialization. This acquires and stores an oauth2 token which remains
        valid for one hour and is used to authenticate all other API requests.

        Raises:
            AmberCloudError: if Amber cloud gives non-200 response
        """

        url = self.license_profile['oauth-server'] + '/oauth2'
        headers = {
            'Content-Type': 'application/json',
            'User-Agent': self.user_agent
        }
        body = {
            'username': self.license_profile['username'],
            'password': self.license_profile['password']
        }

        try:
            response = requests.request(method='POST', url=url, headers=headers, json=body,
                                        verify=self.license_profile['verify'],
                                        cert=self.license_profile['cert'],
                                        timeout=self.timeout)
        except requests.exceptions.Timeout:
            raise AmberCloudError(500, "request timed out")
        except Exception as e:
            raise AmberCloudError(401, 'invalid server connection')
        if response.status_code != 200:
            message = "authentication failed: {}".format(response.json().get('message', 'no message'))
            raise AmberCloudError(response.status_code, message)

        # invalid credentials return a 200 where token is an empty string
        self.token = response.json().get('idToken')
        if self.token is None:
            raise AmberCloudError(401, "authentication failed: invalid credentials")

        expire_secs = int(response.json().get('expiresIn'))
        if expire_secs is None:
            raise AmberCloudError(401, "authentication failed: missing expiration")
        self.reauth_time = time.time() + expire_secs - 60

    def _api_call(self, method, url, headers, body=None):
        """Make a REST call to the Amber server and handle the response"""

        if time.time() > self.reauth_time:
            self._authenticate()

        headers['Authorization'] = 'Bearer {}'.format(self.token)
        headers['User-Agent'] = self.user_agent
        try:
            response = requests.request(method=method, url=url, headers=headers, json=body,
                                        verify=self.license_profile['verify'],
                                        cert=self.license_profile['cert'],
                                        timeout=self.timeout)
        except requests.exceptions.Timeout:
            # request timed out
            raise AmberCloudError(500, "request timed out")

        if response.status_code != 200 and response.status_code != 202:
            raise AmberCloudError(response.status_code, response.json().get('message', 'no message'))

        if 'code' in response.json() and response.json()['code'] != response.status_code:
            # todo: if this happens, there is a bug in the amber service.
            # is returned in the message, it should agree with the header
            raise AmberCloudError(response.json()['code'], response.json().get('message', 'no message'))

        # todo: we should not see errorMessage in the response.  It indicates a misconfigured API gateway
        if 'errorMessage' in response.json():
            raise AmberCloudError(500, response.json()['errorMessage'])

        return response.json()

    def get_version(self):
        """Get version information for Amber

        Returns:
            json version information

        Raises:
            AmberCloudError: if Amber cloud gives non-200 response
        """

        url = self.license_profile['server'] + '/version'
        headers = {
            'Content-Type': 'application/json'
        }
        response = self._api_call('GET', url, headers)

        return response

    def create_sensor(self, label=''):
        """Create a new sensor instance

        Args:
            label (str): label to assign to created sensor

        Returns:
            A string containing the `sensor_id` that was created

        Raises:
            AmberUserError: if client is not authenticated
            AmberCloudError: if Amber cloud gives non-200 response
        """

        url = self.license_profile['server'] + '/sensor'
        headers = {
            'Content-Type': 'application/json'
        }
        body = {
            'label': label
        }
        response = self._api_call('POST', url, headers, body=body)
        sensor_id = response['sensorId']

        return sensor_id

    def update_label(self, sensor_id, label):
        """Update the label of a sensor instance

        Args:
            sensor_id (str): sensor identifier
            label (str): new label to assign to sensor

        Returns:
            A string containing the new label assigned to sensor

        Raises:
            AmberUserError: if client is not authenticated
            AmberCloudError: if Amber cloud gives non-200 response
        """

        url = self.license_profile['server'] + '/sensor'
        headers = {
            'Content-Type': 'application/json',
            'sensorId': sensor_id
        }
        body = {
            'label': label
        }
        response = self._api_call('PUT', url, headers, body=body)
        label = response['label']

        return label

    def delete_sensor(self, sensor_id):
        """Delete an amber sensor instance

        Args:
            sensor_id (str): sensor identifier

        Raises:
            AmberUserError: if client is not authenticated
            AmberCloudError: if Amber cloud gives non-200 response
        """

        url = self.license_profile['server'] + '/sensor'
        headers = {
            'Content-Type': 'application/json',
            'sensorId': sensor_id
        }
        response = self._api_call('DELETE', url, headers)

    def list_sensors(self):
        """List all sensor instances currently associated with Amber account

        Returns:
            A dictionary mapping sensor IDs to corresponding labels

        Raises:
            AmberUserError: if client is not authenticated
            AmberCloudError: if Amber cloud gives non-200 response
        """

        url = self.license_profile['server'] + '/sensors'
        headers = {
            'Content-Type': 'application/json'
        }
        response = self._api_call('GET', url, headers)
        sensors = {s['sensorId']: s.get('label', None) for s in response}

        return sensors

    def configure_sensor(self, sensor_id, feature_count=1, streaming_window_size=25,
                         samples_to_buffer=10000,
                         anomaly_history_window=10000,
                         learning_rate_numerator=10,
                         learning_rate_denominator=10000,
                         learning_max_clusters=1000,
                         learning_max_samples=1000000,
                         features=None):
        """Configure an amber sensor instance

        Args:
            sensor_id (str): sensor identifier
            feature_count (int): number of features (dimensionality of each data sample)
            streaming_window_size (int): streaming window size (number of samples)
            samples_to_buffer (int): number of samples to load before autotuning
            anomaly_history_window (int): number of samples to use for AH calculation
            learning_rate_numerator (int): sensor "graduates" (i.e. transitions from
                learning to monitoring mode) if fewer than learning_rate_numerator
                new clusters are opened in the last learning_rate_denominator samples
            learning_rate_denominator (int): see learning_rate_numerator
            learning_max_clusters (int): sensor graduates if this many clusters are created
            learning_max_samples (int): sensor graduates if this many samples are processed
            features (list): optional list of per feature settings (minVal, maxVal, and label)

        Returns:
            A dictionary containing:

                {
                    'feature_count': int,
                    'streaming_window_size': int,
                    'samples_to_buffer': int
                    'anomaly_history_window': int,
                    'learning_rate_numerator': int,
                    'learning_rate_denominator': int,
                    'learning_max_clusters': int,
                    'learning_max_samples': int,
                    'features': [ {'minVal': float, 'maxVal': float, 'label': string}, {...} ]
                }

        Raises:
            AmberUserError: if client is not authenticated or supplies invalid options
            AmberCloudError: if Amber cloud gives non-200 response
        """
        if features is None:
            features = []

        if not feature_count > 0 or not isinstance(feature_count, Integral):
            raise AmberUserError("invalid 'feature_count': must be positive integer")

        if not streaming_window_size > 0 or not isinstance(streaming_window_size, Integral):
            raise AmberUserError("invalid 'streaming_window_size': must be positive integer")

        url = self.license_profile['server'] + '/config'
        headers = {
            'Content-Type': 'application/json',
            'sensorId': sensor_id
        }
        body = {
            'featureCount': feature_count,
            'streamingWindowSize': streaming_window_size,
            'samplesToBuffer': samples_to_buffer,
            'anomalyHistoryWindow': anomaly_history_window,
            'learningRateNumerator': learning_rate_numerator,
            'learningRateDenominator': learning_rate_denominator,
            'learningMaxClusters': learning_max_clusters,
            'learningMaxSamples': learning_max_samples,
            'features': features
        }
        config = self._api_call('POST', url, headers, body=body)

        return config

    def _isiterable(self, x):
        # consider strings non-iterable for shape validation purposes,
        # that way they are printed out whole when caught as nonnumeric
        if isinstance(x, str):
            return False

        # collections.abc docs: "The only reliable way to determine
        # whether an object is iterable is to call iter(obj)."
        try:
            iter(x)
        except TypeError:
            return False

        return True

    def _validate_dims(self, data):
        """Validate that data is non-empty and one of the following:
           scalar value, list-like or list-of-lists-like where all
           sublists have equal length. Return 0, 1 or 2 as inferred
           number of array dimensions
        """

        # not-iterable data is a single scalar data point
        if not self._isiterable(data):
            return 0

        # iterable and unnested data is a 1-d array
        if not any(self._isiterable(d) for d in data):
            if len(list(data)) == 0:
                raise ValueError("empty")

            return 1

        # iterable and nested data is 2-d array
        if not all(self._isiterable(d) for d in data):
            raise ValueError("cannot mix nested scalars and iterables")

        sublengths = [len(list(d)) for d in data]
        if len(set(sublengths)) > 1:
            raise ValueError("nested sublists must have equal length")

        flattened_2d = list(itertools.chain.from_iterable(data))

        if any(isinstance(i, Iterable) for i in flattened_2d):
            raise ValueError("cannot be nested deeper than list-of-lists")

        if sublengths[0] == 0:
            raise ValueError("empty")

        return 2

    def _convert_to_csv(self, data):
        """Validate data and convert to a comma-separated plaintext string"""

        # Note: as in the Boon Nano SDK, there is no check that data dimensions
        # align with number of features and streaming window size.
        ndim = self._validate_dims(data)

        if ndim == 0:
            data_flat = [data]
        elif ndim == 1:
            data_flat = list(data)
        elif ndim == 2:
            data_flat = list(itertools.chain.from_iterable(data))

        for d in data_flat:
            if not isinstance(d, Number):
                raise ValueError("contained {} which is not numeric".format(d.__repr__()))

        return ','.join([str(float(d)) for d in data_flat])

    def pretrain_sensor(self, sensor_id, data, autotune_config=True, block=True):
        """Pretrain a sensor with historical data

        Args:
            sensor_id (str): sensor identifier
            data (array-like): data to be inferenced. Must be non-empty,
                entirely numeric and one of the following: scalar value,
                list-like or list-of-lists-like where all sublists have
                equal length.
            autotune_config (bool): if True, the sensor will be reconfigured based
                on the training data provided so that the sensor will be in monitoring
                once the data is through. If False, the sensor uses the already
                configured values to train the sensor.
            block (bool): if True, will block until pretraining is complete.
                Otherwise, will return immediately; in this case pretraining
                status can be checked using get_pretrain_state endpoint.

        Returns:

            {
                'state': str
            }

            'state': current state of the sensor.
                "Pretraining": pretraining is in progress

        Raises:
            AmberUserError: if client is not authenticated or supplies invalid data
            AmberCloudError: if Amber cloud gives non-200 response
        """

        # Server expects data as a plaintext string of comma-separated values.
        try:
            data_csv = self._convert_to_csv(data)
        except ValueError as e:
            raise AmberUserError("invalid data: {}".format(e))

        url = self.license_profile['server'] + '/pretrain'
        headers = {
            'Content-Type': 'application/json',
            'sensorId': sensor_id
        }
        body = {
            'data': data_csv,
            'autotuneConfig': autotune_config
        }

        results = self._api_call('POST', url, headers, body=body)

        if not block:
            return results

        while True:
            results = self.get_pretrain_state(sensor_id)
            if results['state'] == "Pretraining":
                time.sleep(5)
                continue
            else:
                return results

    def get_pretrain_state(self, sensor_id):
        """Gets the state of sensor that is being pretrained

        Args:
            sensor_id (str): sensor identifier

        Returns:

            {
                    'state': str
            }

            'state': current state of the sensor. One of:
                "Pretraining": pretraining is in progress
                "Pretrained": pretraining has completed
                "Error": error has occurred
        Raises:
            AmberUserError: if client is not authenticated or supplies invalid data
            AmberCloudError: if Amber cloud gives non-200 response
        """

        url = self.license_profile['server'] + '/pretrain'
        headers = {
            'Content-Type': 'application/json',
            'sensorId': sensor_id
        }

        results = self._api_call('GET', url, headers)

        return results

    def stream_sensor(self, sensor_id, data, save_image=True):
        """Stream data to an amber sensor and return the inference result

        Args:
            sensor_id (str): sensor identifier
            data (array-like): data to be inferenced. Must be non-empty,
                entirely numeric and one of the following: scalar value,
                list-like or list-of-lists-like where all sublists have
                equal length.
            save_image (bool): whether to save the image after calculation

        Returns:
            A dictionary containing inferencing results:

                {
                    'state': str,
                    'message': str,
                    'progress': int,
                    'clusterCount': int,
                    'retryCount': int,
                    'streamingWindowSize': int,
                    'totalInferences': int,
                    'ID': [int],
                    'SI': [int],
                    'AD': [int],
                    'AH': [int],
                    'AM': [float],
                    'AW': [int]
                }

                'state': current state of the sensor. One of:
                    "Buffering": gathering initial sensor data
                    "Autotuning": autotuning configuration in progress
                    "Learning": sensor is active and learning
                    "Monitoring": sensor is active but monitoring only (learning disabled)
                    "Error": fatal error has occurred
                'message': accompanying message for current sensor state
                'progress' progress as a percentage value (applicable for "Buffering" and "Autotuning" states)
                'clusterCount' number of clusters created so far
                'retryCount' number of times autotuning was re-attempted to tune streamingWindowSize
                'streamingWindowSize': streaming window size of sensor (may differ from value
                    given at configuration if window size was adjusted during autotune)
                'totalInferences': number of inferences since configuration
                'ID': list of cluster IDs. The values in this list correspond one-to-one
                    with input samples, indicating the cluster to which each input pattern
                    was assigned.
                'SI': smoothed anomaly index. The values in this list correspond
                    one-for-one with input samples and range between 0 and 1000. Values
                    closer to 0 represent input patterns which are ordinary given the data
                    seen so far on this sensor. Values closer to 1000 represent novel patterns
                    which are anomalous with respect to data seen before.
                'RI': raw anomaly index. These values are the SI values without any smoothing.
                'AD': list of binary anomaly detection values. These correspond one-to-one
                    with input samples and are produced by thresholding the smoothed anomaly
                    index (SI). The threshold is determined automatically from the SI values.
                    A value of 0 indicates that the SI has not exceeded the anomaly detection
                    threshold. A value of 1 indicates it has, signaling an anomaly at the
                    corresponding input sample.
                'AH': list of anomaly history values. These values are a moving-window sum of
                    the AD value, giving the number of anomaly detections (1's) present in the
                    AD signal over a "recent history" window whose length is the buffer size.
                'AM': list of "Amber Metric" values. These are floating point values between
                    0.0 and 1.0 indicating the extent to which each corresponding AH value
                    shows an unusually high number of anomalies in recent history. The values
                    are derived statistically from a Poisson model, with values close to 0.0
                    signaling a lower, and values close to 1.0 signaling a higher, frequency
                    of anomalies than usual.
                'AW': list of "Amber Warning Level" values. This index is produced by thresholding
                    the Amber Metric (AM) and takes on the values 0, 1 or 2 representing a discrete
                    "warning level" for an asset based on the frequency of anomalies within recent
                    history. 0 = normal, 1 = asset changing, 2 = asset critical. The default
                    thresholds for the two warning levels are the standard statistical values
                    of 0.95 (outlier, asset changing) and 0.997 (extreme outlier, asset critical).

        Raises:
            AmberUserError: if client is not authenticated or supplies invalid data
            AmberCloudError: if Amber cloud gives non-200 response
        """

        # Server expects data as a plaintext string of comma-separated values.
        try:
            data_csv = self._convert_to_csv(data)
        except ValueError as e:
            raise AmberUserError("invalid data: {}".format(e))

        url = self.license_profile['server'] + '/stream'
        headers = {
            'Content-Type': 'application/json',
            'sensorId': sensor_id
        }
        body = {
            'saveImage': save_image,
            'data': data_csv
        }

        results = self._api_call('POST', url, headers, body=body)

        return results

    def get_sensor(self, sensor_id):
        """Get info about a sensor

        Args:
            sensor_id (str): sensor identifier

        Returns:
            A dictionary containing sensor information:

                {
                    'label': str,
                    'sensorId': str,
                    'usageInfo': {
                        putSensor {
                            'callsTotal': int
                            'callsThisPeriod': int
                            'lastCalled': str
                        },
                        getSensor {
                            'callsTotal': int
                            'callsThisPeriod': int
                            'lastCalled': str
                        },
                        getConfig {
                            'callsTotal': int
                            'callsThisPeriod': int
                            'lastCalled': str
                        },
                        postStream {
                            'callsTotal': int
                            'callsThisPeriod': int
                            'lastCalled': int
                            'samplesTotal': int
                            'samplesThisPeriod': int
                        }
                        getStatus {
                            'callsTotal': int
                            'callsThisPeriod': int
                            'lastCalled': str
                        }
                    }
                }

                'label' (str): sensor label
                'sensorId' (str): sensor identifier
                'callsTotal': total number of calls to this endpoint
                'callsThisPeriod': calls this billing period to this endpoint
                'lastCalled': ISO formatted time of last call to this endpoint
                'samplesTotal': total number of samples processed
                'samplesThisPeriod': number of samples processed this billing period

        Raises:
            AmberUserError: if client is not authenticated
            AmberCloudError: if Amber cloud gives non-200 response
        """

        url = self.license_profile['server'] + '/sensor'
        headers = {
            'Content-Type': 'application/json',
            'sensorId': sensor_id
        }
        sensor = self._api_call('GET', url, headers)

        return sensor

    def get_config(self, sensor_id):
        """Get current sensor configuration

        Args:
            sensor_id (str): sensor identifier

        Returns:
            A dictionary containing the current sensor configuration:

                {
                    'featureCount': int,
                    'streamingWindowSize': int,
                    'samplesToBuffer': int,
                    'anomalyHistoryWindow': int,
                    'learningRateNumerator': int,
                    'learningRateDenominator': int,
                    'learningMaxClusters': int,
                    'learningMaxSamples': int,
                    'percentVariation': float,
                    'features':
                    [
                        {
                            'min': float,
                            'max': float
                        }
                    ]
                }

                'featureCount': number of features (dimensionality of each data sample)
                'streamingWindowSize': streaming window size (number of samples)
                'samplesToBuffer': number of samples to load before autotuning
                'anomalyHistoryWindow': number of samples to calculate normal anomaly variation
                'learningRateNumerator': sensor "graduates" (i.e. transitions from
                    learning to monitoring mode) if fewer than learning_rate_numerator
                    new clusters are opened in the last learning_rate_denominator samples
                'learningRateDenominator': see learning_rate_numerator
                'learningMaxClusters': sensor graduates if this many clusters are created
                'learningMaxSamples': sensor graduates if this many samples are processed
                'percentVariation': percent variation parameter discovered by autotuning
                'features': min/max values per feature discovered by autotuning
        Raises:
            AmberUserError: if client is not authenticated
            AmberCloudError: if Amber cloud gives non-200 response
        """

        url = self.license_profile['server'] + '/config'
        headers = {
            'Content-Type': 'application/json',
            'sensorId': sensor_id
        }
        config = self._api_call('GET', url, headers)

        return config

    def get_status(self, sensor_id):
        """Get sensor status

        Args:
            sensor_id (str): sensor identifier

        Returns:
            A dictionary containing the clustering status for a sensor:

                {
                    'pca' [(int,int,int)],
                    'clusterGrowth' [int],
                    'clusterSizes' [int],
                    'anomalyIndexes' [int],
                    'frequencyIndexes' [int],
                    'distanceIndexes' [int],
                    'totalInferences' [int],
                    'numClusters' [int],
                }

                'pca': list of length-3 vectors representing cluster centroids
                    with dimensionality reduced to 3 principal components. List length
                    is one plus the maximum cluster ID, with element 0 corresponding
                    to the "zero" cluster, element 1 corresponding to cluster ID 1, etc.
                'clusterGrowth': sample index at which each new cluster was created.
                    Elements for this and other list results are ordered as in 'pca'.
                'clusterSizes': number of samples in each cluster
                'anomalyIndexes': anomaly index associated with each cluster
                'frequencyIndexes': frequency index associated with each cluster
                'distanceIndexes': distance index associated with each cluster
                'totalInferences': total number of inferences performed so far
                'numClusters': number of clusters created so far (includes zero cluster)

        Raises:
            AmberUserError: if client is not authenticated
            AmberCloudError: if Amber cloud gives non-200 response
        """

        url = self.license_profile['server'] + '/status'
        headers = {
            'Content-Type': 'application/json',
            'sensorId': sensor_id
        }
        status = self._api_call('GET', url, headers)

        return status

    def get_root_cause(self, sensor_id, id_list=[], pattern_list=[]):
        """Get root cause

        Args:
            sensor_id (str): sensor identifier
            id_list (list): list of IDs to return the root cause for
            pattern_list (list): list of pattern vectors to calculate the root cause against the model

        Returns:
            A list containing the root cause for each pattern/id provided for a sensor:

                [float]

        Raises:
            AmberUserError: if client is not authenticated
            AmberCloudError: if Amber cloud gives non-200 response
        """

        if len(id_list) != 0 and len(pattern_list) != 0:
            raise AmberUserError('Cannot specify both patterns and cluster IDs for analysis')
        url_call = 'rootCause?'
        if len(id_list) != 0:
            # IDs
            id_list = [str(element) for element in id_list]
            url_call = url_call + 'clusterID=[' + ",".join(id_list) + ']'
        elif len(pattern_list) != 0:
            # patterns
            if len(np.array(pattern_list).shape) == 1:  # only 1 pattern provided
                pattern_list = [pattern_list]
            else:
                for i, pattern in enumerate(pattern_list):
                    pattern_list[i] = ','.join([str(element) for element in pattern])
            url_call = url_call + 'pattern=[[' + "],[".join(pattern_list) + ']]'
        else:
            raise AmberUserError('Must specify either cluster IDs or patterns to analyze')

        url = self.license_profile['server'] + '/' + url_call
        headers = {
            'Content-Type': 'application/json',
            'sensorId': sensor_id
        }
        status = self._api_call('GET', url, headers)

        return status



import csv
import os
import time
from nose.tools import assert_equal
from nose.tools import assert_true
from nose.tools import assert_false
from nose.tools import assert_raises
from nose.tools import assert_not_equal
from nose.tools import assert_is_not_none
from nose.tools import assert_greater
from nose.tools import with_setup
from boonamber import AmberClient, AmberUserError, AmberCloudError
from secrets import get_secrets

saved_license_id = None
license_id = None
license_profile = {}


# secrets downloaded from points beyond
def amber_set_test_profile():
    global license_id, license_profile

    local_license_file = os.environ.get('AMBER_LICENSE_FILE')
    if os.environ.get('AMBER_LICENSE_FILE', None) is not None:
        # load credential set for test using the AMBER_LICENSE_FILE specified
        # in the environment
        amber_client = AmberClient()
        os.environ['AMBER_USERNAME'] = amber_client.license_profile['username']
        os.environ['AMBER_PASSWORD'] = amber_client.license_profile['password']
        os.environ['AMBER_SERVER'] = amber_client.license_profile['server']
        os.environ['AMBER_OAUTH_SERVER'] = amber_client.license_profile['oauth-server']
    else:
        # load credential set from secrets manager, AMBER_LICENSE_ID must be specified in environment
        # to select which credential set to use
        if license_id is None:
            license_id = os.environ.get('AMBER_LICENSE_ID', None)
            assert_is_not_none(license_id, 'AMBER_LICENSE_ID must be specified in environment')
            secret_dict = get_secrets()
            license_profile = secret_dict.get(license_id, None)
            assert_is_not_none(license_profile, 'license_id {} not found'.format(license_id))
        os.environ['AMBER_USERNAME'] = license_profile['username']
        os.environ['AMBER_PASSWORD'] = license_profile['password']
        os.environ['AMBER_SERVER'] = license_profile['server']
        os.environ['AMBER_OAUTH_SERVER'] = license_profile.get('oauth-server', license_profile['server'])


class Test_01_AmberInstance:
    # class variable to saved license file name from environment
    saved_env = {
        'AMBER_LICENSE_FILE': None,
        'AMBER_USERNAME': None,
        'AMBER_PASSWORD': None,
        'AMBER_SERVER': None,
        'AMBER_OAUTH_SERVER': None,
        'AMBER_LICENSE_ID': None,
        'AMBER_SSL_CERT': None,
        'AMBER_SSL_VERIFY': None
    }

    @staticmethod
    def clear_environment():
        for key in Test_01_AmberInstance.saved_env:
            if key in os.environ:
                Test_01_AmberInstance.saved_env[key] = os.environ.get(key, None)
                del os.environ[key]

    @staticmethod
    def restore_environment():
        for key, value in Test_01_AmberInstance.saved_env.items():
            if value is not None:
                os.environ[key] = value
            elif key in os.environ:
                del os.environ[key]

    def test_01_init(self):

        Test_01_AmberInstance.clear_environment()

        # load profile using license file specified as parameter
        profile1 = AmberClient(license_file='test.Amber.license')

        # load the same profile using license loaded from environment
        os.environ['AMBER_LICENSE_FILE'] = 'test.Amber.license'
        profile2 = AmberClient()
        assert_equal(profile1.license_profile, profile2.license_profile, 'override with AMBER_LICENSE_FILE')

        # override items in license file through environment
        os.environ['AMBER_USERNAME'] = "xyyyAmberUser"
        os.environ['AMBER_PASSWORD'] = "bogus_password"
        os.environ['AMBER_SERVER'] = "https://temp.amber.boonlogic.com/v1"
        os.environ['AMBER_SSL_CERT'] = "bogus_ssl_cert"
        os.environ['AMBER_SSL_VERIFY'] = "false"
        profile3 = AmberClient(license_file='test.Amber.license')
        assert_equal(profile3.license_profile['server'], "https://temp.amber.boonlogic.com/v1")
        assert_equal(profile3.license_profile['username'], "xyyyAmberUser")
        assert_equal(profile3.license_profile['password'], "bogus_password")
        assert_equal(profile3.license_profile['cert'], "bogus_ssl_cert")
        assert_equal(profile3.license_profile['verify'], False)

        # set configuration through environment with non-existent license file
        os.environ['AMBER_USERNAME'] = "EnvironmentAmberUser"
        os.environ['AMBER_PASSWORD'] = "bogus_password"
        os.environ['AMBER_SERVER'] = "https://temp.amber.boonlogic.com/v1"
        os.environ['AMBER_SSL_CERT'] = "bogus_ssl_cert"
        os.environ['AMBER_SSL_VERIFY'] = "false"
        profile4 = AmberClient(license_file='bogus.Amber.license')
        assert_equal(profile4.license_profile['server'], "https://temp.amber.boonlogic.com/v1")
        assert_equal(profile4.license_profile['username'], "EnvironmentAmberUser")
        assert_equal(profile4.license_profile['password'], "bogus_password")
        assert_equal(profile4.license_profile['cert'], "bogus_ssl_cert")
        assert_equal(profile4.license_profile['verify'], False)

        Test_01_AmberInstance.restore_environment()

    def test_02_init_negative(self):

        Test_01_AmberInstance.clear_environment()

        # no license file specified
        assert_raises(AmberUserError, AmberClient, license_id="default", license_file="nonexistent-license-file")

        # missing required fields
        os.environ['AMBER_LICENSE_FILE'] = "test.Amber.license"
        assert_raises(AmberUserError, AmberClient, license_id="nonexistent-license-id",
                      license_file="test.Amber.license")
        assert_raises(AmberUserError, AmberClient, license_id="missing-username", license_file="test.Amber.license")
        assert_raises(AmberUserError, AmberClient, license_id="missing-password", license_file="test.Amber.license")
        assert_raises(AmberUserError, AmberClient, license_id="missing-server", license_file="test.Amber.license")

        Test_01_AmberInstance.restore_environment()


class Test_02_Authenticate:

    def test_01_authenticate(self):
        global license_id
        amber_set_test_profile()
        amber = AmberClient(license_file=None, license_id=None)
        amber._authenticate()
        assert_not_equal(amber.token, None)
        assert_not_equal(amber.token, '')

    def test_02_authenticate_negative(self):
        # wrong password
        os.environ['AMBER_PASSWORD'] = "not-valid"
        self.amber = AmberClient(license_file=None, license_id=None)
        with assert_raises(AmberCloudError) as context:
            self.amber._authenticate()
        assert_equal(context.exception.code, 401)
        del os.environ['AMBER_PASSWORD']


class Test_03_SensorOps:
    # class variables
    amber = None
    sensor_id = None

    def test_01_create_sensor(self):

        amber_set_test_profile()

        Test_03_SensorOps.amber = AmberClient(license_file=None, license_id=None)
        try:

            sensor_list = Test_03_SensorOps.amber.list_sensors()
            for sensor, label in sensor_list.items():
                # look for existing test sensors
                if label == 'test-sensor-python':
                    Test_03_SensorOps.sensor_id = sensor
                    break
            if Test_03_SensorOps.sensor_id is None:
                # create a sensor if no test sensors found
                Test_03_SensorOps.sensor_id = Test_03_SensorOps.amber.create_sensor('test-sensor-python')
            assert_not_equal(Test_03_SensorOps.sensor_id, None)
            assert_not_equal(Test_03_SensorOps.sensor_id, "")
        except Exception as e:
            raise RuntimeError("setup failed: {}".format(e))

    def test_02_update_label(self):
        label = Test_03_SensorOps.amber.update_label(Test_03_SensorOps.sensor_id, 'new-label')
        assert_equal(label, 'new-label')

        try:
            Test_03_SensorOps.amber.update_label(Test_03_SensorOps.sensor_id, 'test-sensor-python')
        except Exception as e:
            raise RuntimeError("teardown failed, label was not changed back to 'test-sensor-python': {}".format(e))

    def test_03_update_label_negative(self):
        with assert_raises(AmberCloudError) as context:
            label = Test_03_SensorOps.amber.update_label('nonexistent-sensor-id', 'test-sensor-python')
        assert_equal(context.exception.code, 404)

    def test_04_get_sensor(self):
        sensor = Test_03_SensorOps.amber.get_sensor(Test_03_SensorOps.sensor_id)
        assert_equal(sensor['label'], 'test-sensor-python')
        assert_equal(sensor['sensorId'], Test_03_SensorOps.sensor_id)
        assert_true('usageInfo' in sensor)

    def test_05_get_sensor_negative(self):
        with assert_raises(AmberCloudError) as context:
            sensor = Test_03_SensorOps.amber.get_sensor('nonexistent-sensor-id')
        assert_equal(context.exception.code, 404)

    def test_08_list_sensors(self):
        sensors = Test_03_SensorOps.amber.list_sensors()
        assert_true(Test_03_SensorOps.sensor_id in sensors.keys())

    def test_06_configure_sensor(self):
        # configure sensor with custom features
        expected = {
            'anomalyHistoryWindow': 0,
            'featureCount': 1,
            'streamingWindowSize': 25,
            'samplesToBuffer': 1000,
            'anomalyHistoryWindow': 1000,
            'learningRateNumerator': 10,
            'learningRateDenominator': 10000,
            'learningMaxClusters': 1000,
            'learningMaxSamples': 1000000,
            'features': [{
                'minVal': 1,
                'maxVal': 50,
                'label': 'fancy-label'
            }]
        }
        features = [{
            'minVal': 1,
            'maxVal': 50,
            'label': 'fancy-label'
        }]
        config = Test_03_SensorOps.amber.configure_sensor(Test_03_SensorOps.sensor_id, feature_count=1,
                                                          streaming_window_size=25,
                                                          samples_to_buffer=1000,
                                                          anomaly_history_window=1000,
                                                          learning_rate_numerator=10,
                                                          learning_rate_denominator=10000,
                                                          learning_max_clusters=1000,
                                                          learning_max_samples=1000000,
                                                          features=features)

        assert_equal(config, expected)

        # configure sensor with default features
        expected = {
            'anomalyHistoryWindow': 0,
            'featureCount': 1,
            'streamingWindowSize': 25,
            'samplesToBuffer': 1000,
            'anomalyHistoryWindow': 1000,
            'learningRateNumerator': 10,
            'learningRateDenominator': 10000,
            'learningMaxClusters': 1000,
            'learningMaxSamples': 1000000,
            'features': [{
                'minVal': 0,
                'maxVal': 1,
                'label': 'feature-0'
            }]
        }
        config = Test_03_SensorOps.amber.configure_sensor(Test_03_SensorOps.sensor_id, feature_count=1,
                                                          streaming_window_size=25,
                                                          samples_to_buffer=1000,
                                                          anomaly_history_window=1000,
                                                          learning_rate_numerator=10,
                                                          learning_rate_denominator=10000,
                                                          learning_max_clusters=1000,
                                                          learning_max_samples=1000000)
        assert_equal(config, expected)

    def test_07_configure_sensor_negative(self):
        with assert_raises(AmberCloudError) as context:
            config = Test_03_SensorOps.amber.configure_sensor('nonexistent-sensor-id')
        assert_equal(context.exception.code, 404)

        # invalid feature_count or streaming_window_size
        assert_raises(AmberUserError, Test_03_SensorOps.amber.configure_sensor, Test_03_SensorOps.sensor_id,
                      feature_count=-1)
        assert_raises(AmberUserError, Test_03_SensorOps.amber.configure_sensor, Test_03_SensorOps.sensor_id,
                      feature_count=1.5)
        assert_raises(AmberUserError, Test_03_SensorOps.amber.configure_sensor, Test_03_SensorOps.sensor_id,
                      streaming_window_size=-1)
        assert_raises(AmberUserError, Test_03_SensorOps.amber.configure_sensor, Test_03_SensorOps.sensor_id,
                      streaming_window_size=1.5)

    def test_08_get_config(self):
        expected = {
            'anomalyHistoryWindow': 0,
            'featureCount': 1,
            'streamingWindowSize': 25,
            'samplesToBuffer': 1000,
            'anomalyHistoryWindow': 1000,
            'learningRateNumerator': 10,
            'learningRateDenominator': 10000,
            'learningMaxClusters': 1000,
            'learningMaxSamples': 1000000,
            'percentVariation': 0.05,
            'features': [{'minVal': 0, 'maxVal': 1, 'label': 'feature-0'}]
        }
        config = Test_03_SensorOps.amber.get_config(Test_03_SensorOps.sensor_id)
        assert_equal(config, expected)

    def test_09_get_config_negative(self):
        with assert_raises(AmberCloudError) as context:
            config = Test_03_SensorOps.amber.get_config('nonexistent-sensor-id')
        assert_equal(context.exception.code, 404)

    def test_10_stream_sensor(self):
        results = Test_03_SensorOps.amber.stream_sensor(Test_03_SensorOps.sensor_id, 1)
        assert_true('state' in results)
        assert_true('message' in results)
        assert_true('progress' in results)
        assert_true('clusterCount' in results)
        assert_true('retryCount' in results)
        assert_true('streamingWindowSize' in results)
        assert_true('SI' in results)
        assert_true('AD' in results)
        assert_true('AH' in results)
        assert_true('AM' in results)
        assert_true('AW' in results)

        # scalar data should return SI of length 1
        assert_true(len(results['SI']) == 1)

        # array data should return SI of same length
        results = Test_03_SensorOps.amber.stream_sensor(Test_03_SensorOps.sensor_id, [1, 2, 3, 4, 5])
        assert_true(len(results['SI']) == 5)

    def test_11_stream_sensor_negative(self):
        with assert_raises(AmberCloudError) as context:
            results = Test_03_SensorOps.amber.stream_sensor('nonexistent-sensor-id', [1, 2, 3, 4, 5])
        assert_equal(context.exception.code, 404)

        # invalid data
        assert_raises(AmberUserError, Test_03_SensorOps.amber.stream_sensor, Test_03_SensorOps.sensor_id, [])
        assert_raises(AmberUserError, Test_03_SensorOps.amber.stream_sensor, Test_03_SensorOps.sensor_id, [1, '2', 3])
        assert_raises(AmberUserError, Test_03_SensorOps.amber.stream_sensor, Test_03_SensorOps.sensor_id,
                      [1, [2, 3], 4])

    def test_12_get_root_cause(self):
        config = Test_03_SensorOps.amber.get_config(Test_03_SensorOps.sensor_id)
        expected = [[0] * len(config['features']) * config['streamingWindowSize']] * 2
        config = Test_03_SensorOps.amber.get_root_cause(Test_03_SensorOps.sensor_id,
                                                        pattern_list=[
                                                            [1] * len(config['features']) * config[
                                                                'streamingWindowSize'],
                                                            [0] * len(config['features']) * config[
                                                                'streamingWindowSize']])
        assert_equal(config, expected)

    def test_13_get_root_cause_negative(self):
        with assert_raises(AmberCloudError) as context:
            config = Test_03_SensorOps.amber.get_root_cause('nonexistent-sensor-id', id_list=[1])
        assert_equal(context.exception.code, 404)

        # give both fail
        with assert_raises(AmberUserError) as context:
            config = Test_03_SensorOps.amber.get_root_cause(Test_03_SensorOps.sensor_id, id_list=[1],
                                                            pattern_list=[[1, 2, 3], [4, 5, 6]])

        # give neither fail
        with assert_raises(AmberUserError) as context:
            config = Test_03_SensorOps.amber.get_root_cause(Test_03_SensorOps.sensor_id)

        assert_raises(AmberCloudError, Test_03_SensorOps.amber.get_root_cause, Test_03_SensorOps.sensor_id, [1])

    def test_14_get_status(self):
        status = Test_03_SensorOps.amber.get_status(Test_03_SensorOps.sensor_id)
        assert_true('pca' in status)
        assert_true('numClusters' in status)

    def test_15_get_status_negative(self):
        with assert_raises(AmberCloudError) as context:
            status = Test_03_SensorOps.amber.get_status('nonexistent-sensor-id')
        assert_equal(context.exception.code, 404)

    def test_16_get_pretrain_state(self):
        response = Test_03_SensorOps.amber.get_pretrain_state(Test_03_SensorOps.sensor_id)
        assert_true('state' in response)
        assert_equal(response['state'], 'None')

    def test_17_get_pretrain_state_negative(self):
        with assert_raises(AmberCloudError) as context:
            response = Test_03_SensorOps.amber.get_pretrain_state('nonexistent-sensor-id')
        assert_equal(context.exception.code, 404)

    def test_18_pretrain_sensor(self):
        with open('output_current.csv', 'r') as f:
            csv_reader = csv.reader(f, delimiter=',')
            data = []
            for row in csv_reader:
                for d in row:
                    data.append(float(d))

        results = Test_03_SensorOps.amber.pretrain_sensor(Test_03_SensorOps.sensor_id, data, block=True)
        assert_equal(results['state'], 'Pretrained')

        results = Test_03_SensorOps.amber.pretrain_sensor(Test_03_SensorOps.sensor_id, data, block=False)
        assert_true('Pretraining' in results['state'] or 'Pretrained' in results['state'])
        while True:
            time.sleep(5)
            results = Test_03_SensorOps.amber.get_pretrain_state(Test_03_SensorOps.sensor_id)
            if results['state'] == 'Pretraining':
                continue
            else:
                break
        assert_equal(results['state'], 'Pretrained')

    def test_19_pretrain_sensor_negative(self):
        with assert_raises(AmberCloudError) as context:
            response = Test_03_SensorOps.amber.pretrain_sensor('nonexistent-sensor-id', [1, 2, 3, 4, 5], block=True)
        assert_equal(context.exception.code, 404)

        # not enough data to fill sample buffer
        with assert_raises(AmberCloudError) as context:
            response = Test_03_SensorOps.amber.pretrain_sensor(Test_03_SensorOps.sensor_id, [1, 2, 3, 4, 5], block=True)
        assert_equal(context.exception.code, 400)

    def test_20_delete_sensor_negative(self):
        with assert_raises(AmberCloudError) as context:
            Test_03_SensorOps.amber.delete_sensor('nonexistent-sensor-id')
        assert_equal(context.exception.code, 404)

    def test_21_delete_sensor(self):
        try:
            Test_03_SensorOps.amber.delete_sensor(Test_03_SensorOps.sensor_id)
        except Exception as e:
            raise RuntimeError("teardown failed, sensor was not deleted: {}".format(e))


class Test_04_ApiReauth:

    def test_api_reauth(self):
        amber_set_test_profile()

        # create amber instance and mark auth_time
        amber = AmberClient(license_id=None, license_file=None)
        saved_reauth_time = amber.reauth_time

        # first call covers reauth case, reauth time should be set bigger than initial time
        _ = amber.list_sensors()
        assert_greater(amber.reauth_time, saved_reauth_time)
        saved_reauth_time = amber.reauth_time

        _ = amber.list_sensors()
        # The reauth time should not have changed
        assert_equal(amber.reauth_time, saved_reauth_time)

        # Add 60 to the reauth_time and reissue api call, reauth should occur
        amber.reauth_time += 61
        _ = amber.list_sensors()
        assert_greater(amber.reauth_time, saved_reauth_time)


class Test_04_CSVConvert:

    def test_convert_to_csv(self):
        amber_set_test_profile()

        amber = AmberClient(license_id=None, license_file=None)

        # valid scalar inputs
        assert_equal("1.0", amber._convert_to_csv(1))
        assert_equal("1.0", amber._convert_to_csv(1.0))

        # valid 1d inputs
        assert_equal("1.0,2.0,3.0", amber._convert_to_csv([1, 2, 3]))
        assert_equal("1.0,2.0,3.0", amber._convert_to_csv([1, 2, 3.0]))
        assert_equal("1.0,2.0,3.0", amber._convert_to_csv([1.0, 2.0, 3.0]))

        # valid 2d inputs
        assert_equal("1.0,2.0,3.0,4.0", amber._convert_to_csv([[1, 2], [3, 4]]))
        assert_equal("1.0,2.0,3.0,4.0", amber._convert_to_csv([[1, 2, 3, 4]]))
        assert_equal("1.0,2.0,3.0,4.0", amber._convert_to_csv([[1], [2], [3], [4]]))
        assert_equal("1.0,2.0,3.0,4.0", amber._convert_to_csv([[1, 2], [3, 4.0]]))
        assert_equal("1.0,2.0,3.0,4.0", amber._convert_to_csv([[1.0, 2.0], [3.0, 4.0]]))

    def test_convert_to_csv_negative(self):
        amber = AmberClient(license_id=None, license_file=None)

        # empty data
        assert_raises(ValueError, amber._convert_to_csv, [])
        assert_raises(ValueError, amber._convert_to_csv, [[]])
        assert_raises(ValueError, amber._convert_to_csv, [[], []])

        # non-numeric data
        assert_raises(ValueError, amber._convert_to_csv, None)
        assert_raises(ValueError, amber._convert_to_csv, 'a')
        assert_raises(ValueError, amber._convert_to_csv, 'abc')
        assert_raises(ValueError, amber._convert_to_csv, [1, None, 3])
        assert_raises(ValueError, amber._convert_to_csv, [1, 'a', 3])
        assert_raises(ValueError, amber._convert_to_csv, [1, 'abc', 3])
        assert_raises(ValueError, amber._convert_to_csv, [[1, None], [3, 4]])
        assert_raises(ValueError, amber._convert_to_csv, [[1, 'a'], [3, 4]])
        assert_raises(ValueError, amber._convert_to_csv, [[1, 'abc'], [3, 4]])

        # badly-shaped data
        assert_raises(ValueError, amber._convert_to_csv, [1, [2, 3], 4])  # mixed nesting
        assert_raises(ValueError, amber._convert_to_csv, [[1, 2], [3, 4, 5]])  # ragged array
        assert_raises(ValueError, amber._convert_to_csv, [[[1, 2, 3, 4]]])  # nested too deep
        assert_raises(ValueError, amber._convert_to_csv, [[[1], [2], [3], [4]]])


class Test_05_Version:

    def test_01_version(self):
        amber = AmberClient(license_id=None, license_file=None)
        version = amber.get_version()
        assert_equal(7, len(version.keys()))
        assert_true('api-version' in version.keys())
        assert_true('builder' in version.keys())
        assert_true('expert-api' in version.keys())
        assert_true('expert-common' in version.keys())
        assert_true('nano-secure' in version.keys())
        assert_true('release' in version.keys())
        assert_true('swagger-ui' in version.keys())
        assert_false('rXelXeXaseX' in version.keys())



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='boonamber',
    version='1.0.25',
    author="BoonLogic",
    author_email="amber-support@boonlogic.com",
    packages=['boonamber'],
    install_requires=['requests', 'numpy'],
    description="An SDK for Boon Amber sensor analytics",
    long_description=long_description,
    license='MIT',
    long_description_content_type="text/markdown",
    url="https://github.com/boonlogic/amber-python-sdk",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent"
    ]
)
